import os
import threading
import datetime
import time
import json
import copy

def _serialize_datetime(obj): 
    ''' Used as an argument for json.dumps() serial conversions within 
    ErrorMonitor and OrderMonitor for human-readable logging: 
    json.dumps(<data>, default=_format_values). Attempts to convert eligible 
    python datetime.datetime datatypes to RCF3339 formatted strings (does NOT 
    throw errors on any failed datetime conversions, json.dumps() will default 
    to making a "best-guess").
    
    Supported conversions:
            datetime.datetime

    Parameters
    ----------
    obj : unk
        Potential datetime value to encode.
    
    Returns
    -------
    None 
        
    '''
    # if datetime object, format as RCF3339 time
    if isinstance(obj, datetime.datetime): 
        return obj.isoformat().replace("+00:00", "Z")
    
    else:
        # json docs requests a type error in place of original value (just
        # continues on with conversions)
        raise TypeError("Type not serializable") 

''' SESSION MONITORS '''
class ErrorMonitor(threading.Thread):
    ''' Monitors a list of Oanda "endpoint interface" objects for client-server
        errors.

    Attributes
    ----------
    `logs` : list[dict]
        A continuously updated list of client-server errors made since the 
        ErrorMonitor() object was started. *Note* If `logPath` is specified on 
        initialization, any pre-existing logs within the specified log file will 
        be loaded into this variable prior to appending additional logs to it.

        Single Log Entry Format:\n
        {
            "entryID" : the log's serialized entry number (beginning at 0)
            "datetime" : the date/time the error was logged (UTC)
            "originClass" : the "endpoint interface" type (a class within `session` module)
                that received the error

            "errorDetails" : the error message details, with contents:
                {
                    "url" : full url that the request was sent to
                    "headers" : headers included in the request (Authorization token is 
                        stripped for security, check independently if this is the 
                        suspected issue)
                    "parameters" : parameters included in the request
                    "payload" : the request's json payload (`None` if "GET" request)
                    "code" : the error status code - will be HTTP code or "other" (999)
                    "message" : the error message received
                }
        }


    `_endpoints` : list
        The list of client "endpoint interface" objects that are currently
        being monitored.

    `_flags` : list[dict]
        A list of flags and recent error messages aligned with each endpoint
        being monitored. Each endpoint's flags and recent message are used to 
        prevent repeatedly recording the same error multiple times.

    `_logPath` : str | None
        Full path to log file, if specified. The log's on-disk format
        will be a json for each log entry (appended with a trailing comma). The
        final trailing comma is stripped prior to re-loading the log into memory.
        [Default=None]

    `_printErrors` : bool
        Whether printing errors to stdout. [Default=False]

    `_stop_signal` : threading.Event
        When set, stops the given thread.

    Methods
    --------------
    `stop()` : func
        Sets `self._stop_signal`, stopping the given thread.

    `start()` : func
        Starts the thread (using `run()` function details).

    `run()` : func
        Overrides threading.Thread.run(). Contiuously iterates over all "endpoint 
        interface" objects, checking to see if any errors have occured between
        the client (trader) and server (Oanda). Logs any errors that occur (no 
        error handling is done). Use `start()` to begin thread, not this function.

    '''

    def __init__(self, 
                 endpoints : list,
                 printErrors : bool = False, 
                 logPath : str | None = None) -> None:
        ''' Initializes ErrorMonitor object.
        
        Parameters
        ----------
        `endpoints` : list
            The list of client "endpoint interface" objects to monitor:

            Supported Object Types:\n
                `session.Account()`\n
                `session.Instruments()`\n
                `session.Orders()`\n
                `session.Trades()`\n
                `session.Positions()`\n
                `session.Transactions()`\n
                `session.Pricing()`

        `printErrors` : bool = False
            Whether to print errors to stdout. [Default=False]

        `logPath` : str | None = None
            (Optional) Full path to log file on disk for recording errors. If
            provided, will attempt to load any pre-existing logs to memory
            before error logging begins. [Default=None]

        Returns
        -------
        `None`

        '''

        # inherit Parent class Thread
        threading.Thread.__init__(self)

        # create a way to stop the thread if need be. 
        self._stop_signal = threading.Event()

        # set endpoints
        self._endpoints = [endpoint for endpoint in endpoints]
        
        # set corresponding endpoint flags
        self._flags = [{"flag" : 0, "message" : {}} for i in range(0, len(self._endpoints))]

        # set parameters
        self._logPath = logPath
        self._printErrors = printErrors

        # load prior logs (if available)
        if logPath:
            
            # (will be created once logging begins if file doesn't exist)
            if os.path.exists(logPath):

                with open(logPath, "r") as oldLogs:
                    # read in all but last "," (logs entries are recorded as "{},")
                    tempLogs = "[" + oldLogs.read()[:-1] + "]"
                    self.logs = json.loads(tempLogs)
        
        # otherwise prep container
        else:
            self.logs = []

        return None

    def stop(self):
        ''' Stops the given monitor thread (prevents continued error logging).
        Parameters
        ----------------
        None

        Returns
        -----------
        None
        '''

        self._stop_signal.set()
        
        return None

    def run(self):
        ''' Overrides threading.Thread.run(). Contiuously iterates over all "endpoint 
        interface" objects, checking to see if any errors have occured between
        the client (trader) and server (Oanda). Logs any errors that occur (no 
        error handling is done). Use `start()` to begin thread, not this.
        
        Parameters
        ----------------
        None

        Returns
        -----------
        `None`

        '''

        # continue until thread is stopped with `self.stop()`
        while not self._stop_signal.is_set():

            # iterate over provided endpoints
            for i in range(0, len(self._endpoints)):

                # aquire error lock to ensure both return code and message are aligned
                self._endpoints[i]._errorLock.acquire()

                # if no error code present, ensure corresponding endpoint flag is properly set
                if (self._endpoints[i].rcode == 200) or \
                   (self._endpoints[i].rcode == 201) or \
                   (self._endpoints[i].rcode == 888):
                    
                    if self._flags[i]["flag"] != 0:
                        self._flags[i]["flag"] = 0
                
                # otherwise, check to see if error needs to be recorded
                else:

                    # if corresponding flag and message already set, don't record error again
                    if (self._flags[i]["flag"] == 1) and \
                       (self._flags[i]["message"] == self._endpoints[i].errorMessage):
                        pass

                    # otherwise, record the new error
                    else:
                        errorEntry = {
                                        "entryID" : len(self.logs),
                                        "datetime" : datetime.datetime.now(datetime.UTC).isoformat().replace("+00:00", "Z"),
                                        "originClass" : self._endpoints[i].__class__.__name__,
                                        "errorDetails" : copy.deepcopy(self._endpoints[i].errorMessage)
                                     }

                        # append new entry in memory
                        self.logs.append(errorEntry)

                        # print error to stdout (if applicable)
                        if self._printErrors:
                            print(json.dumps(errorEntry, default=_serialize_datetime, indent=4))

                        # record new entry to log file (if applicable)
                        if self._logPath:
                            # creates file if path doesn't exist
                            with open(self._logPath, "a+") as logFile:
                                logFile.write(json.dumps(errorEntry, default=_serialize_datetime, indent=4) + ",")

                        # set corresponding error flag & message
                        self._flags[i]["flag"] = 1
                        self._flags[i]["message"] = errorEntry["errorDetails"]

                # release error lock
                self._endpoints[i]._errorLock.release()

        return None

class StreamMonitor(threading.Thread):
    ''' Monitors "endpoint interface" streams for their respective heartbeats,
    setting "dead stream" error messages and restarting streams as needed.
    
    Attributes
    ----------
    `deadOnArrival` : datetime.timedelta
        "Dead On Arrival" - number of seconds between heartbeats before a stream
        is considered dead. Oanda heartbeats are ~5s. Stream re-starts take ~2s.
        Allow at least ~7s (preferably more) if modifying this attribute. [Default=10]

    `doNotResusitate` : int
        "Do Not Resuscitate" - number of times to attempt to restart a stream.
        [Default=3]

    `resetTime` : datetime.timedelta
        Number of minutes before resetting `doNotResusitate` counters back to zero for each 
        endpoint. *Note* `doNotResusitate` is meant to give time for minor issues to resolve 
        themselves (brief drops in WiFi service, ISP routing issues, etc.) - 
        `resetTime` gives time for higher-level issues to be resolved (Example: 
        the Oanda streaming server crashes, but is brought back up by their 
        IT department within the hour). [Default=60]

    `_endpoints` : list
        The list of client "endpoint interface" objects with potential streams.

    `_dnrCount` : list
        A list with the number of times each endpoint has had its stream restarted
        since StreamMonitor() began, aligned with `self._endpoints`.

    `_lastReset` : datetime.datetime
        The last time all stream-reset attempts (`_dnrCount`) were set back to zero.

    Methods
    -------
    `stop()` : func
        Sets `self._stop_signal`, stopping the given thread.

    `start()` : func
        Starts the thread (using `run()` function details).

    `run()` : func
        Overrides threading.Thread.run(). Contiuously iterates over all "endpoint 
        interface" objects with streams, checking to see if any of the streams'
        heartbeats are timestamped later than `self.deadOnArrival` minimum number of
        seconds (restarts them if they are). Use `start()` to begin thread, not 
        this function.
    
    '''

    def __init__(self,
                 endpoints : list,
                 deadOnArrival : int = 10,
                 doNotResusitate : int = 3,
                 resetTime : int = 60) -> None:
        ''' Initializes StreamMonitor object.

        Parameters
        ----------
        `endpoints` : list
            The list of client "endpoint interface" objects with potential streams
            to monitor.

            Supported Endpoint Classes:\n
            `sessions.Transactions()`\n
            `sessions.Pricing()`
        
        `deadOnArrival` : int = 10
            "Dead On Arrival" - number of seconds between heartbeats before 
            a stream is considered dead. [Default=10]

        `doNotResusitate` : int = 3
            "Do Not Resuscitate" - number of times to attempt to restart 
            a dead stream. [Default=3]

        `resetTime` : int = 60
            Number of minutes before resetting `doNotResusitate` counters back to zero for each 
            endpoint. [Default=60]

        Returns
        -------
        `None`
        
        '''

        # inherit Parent attributes
        threading.Thread.__init__(self)
        
        # enable way to stop the thread
        self._stop_signal = threading.Event()

        # load attributes
        self.deadOnArrival = datetime.timedelta(seconds=deadOnArrival)
        self.doNotResusitate = doNotResusitate
        self.resetTime = datetime.timedelta(minutes=resetTime)
        self._endpoints = [endpoint for endpoint in endpoints]
        self._dnrCount = [0 for i in range(0, len(self._endpoints))]
        self._lastReset = datetime.datetime.now(datetime.UTC)

        return None

    def stop(self):
        ''' Stops the given monitor thread (prevents continued stream management).
        Parameters
        ----------------
        None

        Returns
        -----------
        None
        '''

        self._stop_signal.set()
        
        return None

    def run(self):
        ''' Overrides threading.Thread.run(). Contiuously iterates over all "endpoint 
        interface" objects with streams, checking to see if any of the streams'
        heartbeats are timestamped later than `self.deadOnArrival` minimum number of
        seconds (restarts them if they are). Use `start()` to begin thread, not 
        this function.

        Parameters
        ----------
        None

        Returns
        -------
        `None`

        '''
        
        # continue until stop flag set
        while not self._stop_signal.is_set():

            # reset stream re-start attempts if enough time (self.resetTime) has passed
            if self._lastReset < datetime.datetime.now(datetime.UTC) - self.resetTime:
                self._dnrCount = [0] * len(self._endpoints)
                self._lastReset = datetime.datetime.now(datetime.UTC)

            # iterate over endpoints
            for i in range(0, len(self._endpoints)):
                
                # if a stream has been started
                if self._endpoints[i].streamHeartbeat:

                    streamIsDead = False
                    
                    # make sure the stream wasn't JUST started
                    if len(self._endpoints[i].streamHeartbeat) != 0:

                        # set deadOnArrival benchmark
                        benchmark = datetime.datetime.now(datetime.UTC) - self.deadOnArrival

                        # if stream is older than deadOnArrival time, stream is considered dead
                        if self._endpoints[i].streamHeartbeat[-1]["time"] < benchmark:
                            streamIsDead = True
                        
                        # otherwise, it's considered fine
                        else:
                            pass

                    # decide what to do with dead stream
                    if streamIsDead:

                        restartStream = False

                        # set message for stream restart attempts
                        if self._dnrCount[i] >= self.doNotResusitate:
                            status_message = "Max stream restarts attempted, leaving stream dead."
                        else:
                            status_message = "Attempting to restart stream..."
                            restartStream = True

                        # acquire lock to prevent logging until message complete
                        self._endpoints[i]._errorLock.acquire()

                        # set error code errorMessage attributes
                        self._endpoints[i].rcode = 999
                        self._endpoints[i].errorMessage = \
                        {
                            "url" : None,
                            "headers" : None,
                            "parameters" : None,
                            "payload" : None, 
                            "code" : 999,
                            "message" : "Dead stream. Prior restart attempts: {}. Restart attempts permitted: {}. {}".format(self._dnrCount[i],
                                                                                                                             self.doNotResusitate,
                                                                                                                             status_message)
                        }

                        # allow error logging
                        self._endpoints[i]._errorLock.release()

                        # restart the stream
                        if restartStream:
                            # ensure old stream stopped
                            self._endpoints[i].stop_stream()
                            time.sleep(2)

                            # start new strem
                            self._endpoints[i].start_stream(*self._endpoints[i]._streamArgs)
                            
                            # give time for stream to start & avoid Oanda request throttling
                            time.sleep(2)
                            
                            self._dnrCount[i] = self._dnrCount[i] + 1

        return None

class OrderMonitor(threading.Thread):
    ''' Monitors a list of Oanda "endpoint interface" objects for successful
    order confirmations.

    Attributes
    ----------
    `logs` : list[dict]
        A continuously updated list of order confirmations made since the 
        OrderMonitor() object was started. *Note* If `logPath` is specified on 
        initialization, any pre-existing logs within the specified log file will 
        be loaded into this variable prior to appending additional logs to it.

        Single Log Entry Format:\n
        {
            "entryID" : the log's serialized entry number (beginning at 0)
            "datetime" : the date/time the order was confirmed (UTC)
            "originClass" : the "endpoint interface" type (a class within `session` module)
                that sent the order
            "confirmationDetails" : Oanda's reponse to successfully receiving the order
        }


    `_endpoints` : list
        The list of client "endpoint interface" objects that are currently
        being monitored.

    `_flags` : list[dict]
        A list of flags and recent confirmation messages aligned with each endpoint
        being monitored. Each endpoint's flags and recent message are used to 
        prevent repeatedly recording the same message multiple times.

    `_logPath` : str | None
        Full path to log file, if specified. The log's on-disk format
        will be a json for each log entry (appended with a trailing comma). The
        final trailing comma is stripped prior to re-loading the log into memory.
        [Default=None]

    `_printConfirmation` : bool
        Whether printing order confirmations to stdout. [Default=False]

    `_stop_signal` : threading.Event
        When set, stops the given thread.

    Methods
    --------------
    `stop()` : func
        Sets `self._stop_signal`, stopping the given thread.

    `start()` : func
        Starts the thread (using `run()` function details).

    `run()` : func
        Overrides threading.Thread.run(). Contiuously iterates over all "endpoint 
        interface" objects, checking to see if any new order confirmation have
        been received, logging them as appropriate. Use `start()` to 
        begin thread, not this function.

    '''

    def __init__(self, 
                 endpoints : list,
                 printConfirmations : bool = False, 
                 logPath : str | None = None) -> None:
        ''' Initializes OrderMonitor object.
        
        Parameters
        ----------
        `endpoints` : list
            The list of client "endpoint interface" objects to monitor:

            Supported Object Types:\n
                `session.Orders()`\n
                `session.Trades()`\n
                `session.Positions()`\n

        `printConfirmations` : bool = False
            Whether to print order confirmations to stdout. [Default=False]

        `logPath` : str | None = None
            (Optional) Full path to log file on disk for recording confirmations. 
            If provided, will attempt to load any pre-existing logs to memory
            before confirmation logging begins. [Default=None]

        Returns
        -------
        `None`

        '''

        # inherit Parent class Thread
        threading.Thread.__init__(self)

        # create a way to stop the thread if need be. 
        self._stop_signal = threading.Event()

        # set endpoints
        self._endpoints = [endpoint for endpoint in endpoints]
        
        # set corresponding endpoint flags
        self._flags = [{"flag" : 0, "message" : {}} for i in range(0, len(self._endpoints))]

        # set parameters
        self._logPath = logPath
        self._printConfirmations = printConfirmations

        # load prior logs (if available)
        if logPath:
            
            # (will be created once logging begins if file doesn't exist)
            if os.path.exists(logPath):

                with open(logPath, "r") as oldLogs:
                    # read in all but last "," (logs entries are recorded as "{},")
                    tempLogs = "[" + oldLogs.read()[:-1] + "]"
                    self.logs = json.loads(tempLogs)
        
        # otherwise prep container
        else:
            self.logs = []

        return None

    def stop(self):
        ''' Stops the given monitor thread (prevents continued order confirmation 
        logging).
        
        Parameters
        ----------------
        None

        Returns
        -----------
        None
        '''

        self._stop_signal.set()
        
        return None

    def run(self):
        ''' Overrides threading.Thread.run(). Contiuously iterates over all "endpoint 
        interface" objects, checking to see if any new order confirmation have
        been received, logging them as appropriate. Use `start()` to 
        begin thread, not this function.
        
        Parameters
        ----------------
        None

        Returns
        -----------
        `None`

        '''

        # continue until thread is stopped with `self.stop()`
        while not self._stop_signal.is_set():

            # iterate over provided endpoints
            for i in range(0, len(self._endpoints)):

                # aquire order lock to ensure both return code and message are aligned
                self._endpoints[i]._orderLock.acquire()

                # if no confirmation code present, ensure flag is properly set
                if (self._endpoints[i].rcode != 888):
                    if self._flags[i]["flag"] != 0:
                        self._flags[i]["flag"] = 0
                
                # otherwise, check to see if confirmation needs to be recorded
                else:
                    
                    # if corresponding flag and message already set, don't record confirmation again
                    if (self._flags[i]["flag"] == 1) and \
                       (self._flags[i]["message"] == self._endpoints[i].orderMessage):
                        pass

                    # otherwise, record the new confirmation
                    else:
                        confirmationEntry = {
                                        "entryID" : len(self.logs),
                                        "datetime" : datetime.datetime.now(datetime.UTC).isoformat().replace("+00:00", "Z"),
                                        "originClass" : self._endpoints[i].__class__.__name__,
                                        "confirmationDetails" : copy.deepcopy(self._endpoints[i].orderMessage)
                                     }

                        # append new entry in memory
                        self.logs.append(confirmationEntry)

                        # print confirmation to stdout (if applicable)
                        if self._printConfirmations:
                            print(json.dumps(confirmationEntry, default=_serialize_datetime, indent=4))

                        # record new entry to log file (if applicable)
                        if self._logPath:
                            # creates file if path doesn't exist
                            with open(self._logPath, "a+") as logFile:
                                logFile.write(json.dumps(confirmationEntry, default=_serialize_datetime, indent=4) + ",")

                        # set corresponding confirmation flag & message
                        self._flags[i]["flag"] = 1
                        self._flags[i]["message"] = confirmationEntry["confirmationDetails"]

                # release order lock
                self._endpoints[i]._orderLock.release()

        return None




from .modules import session
from .modules import monitor
from .modules.prebuild import MarketOrder, LimitOrder, StopOrder, \
                             MarketIfTouchedOrder, TakeProfitOrder, \
                             StopLossOrder, GuaranteedStopLossOrder, \
                             TrailingStopLossOrder


class OandaSession():
    ''' Wrapper class for all sub-session and monitor classes - used to create a 
    single object ("session") with access to all Oanda endpoints / sub-API calls.
    
    Attributes
    ----------
    `account` : modules.session.Account()
        Oanda account API:
        https://developer.oanda.com/rest-live-v20/account-ep/

    `instruments` : modules.session.Instruments()
        Oanda instrument API:
        https://developer.oanda.com/rest-live-v20/instrument-ep/
    
    `orders` : modules.session.Orders()
        Oanda orders API:
        https://developer.oanda.com/rest-live-v20/order-ep/
    
    `trades` : modules.session.Trades()
        Oanda trades API:
        https://developer.oanda.com/rest-live-v20/trade-ep/
    
    `positions` : modules.session.Positions()
        Oanda positions API:
        https://developer.oanda.com/rest-live-v20/position-ep/
    
    `transactions` : modules.session.Transactions()
        Oanda transactions API:
        https://developer.oanda.com/rest-live-v20/transaction-ep/
    
    `pricing` : modules.session.Pricing()
        Oanda pricing API:
        https://developer.oanda.com/rest-live-v20/pricing-ep/
    
    `errorMonitor` : modules.monitor.ErrorMonitor()
        Monitors Oanda interfaces for client-server errors.
    
    `streamMonitor` : modules.monitor.StreamMonitor()
        Monitors Oanda interface streams for connection interuptions, setting 
        stream alert error messages and restarting streams as needed.
    
    `orderMonitor` : modules.monitor.OrderMonitor()
        Monitors Oanda interfaces for successful order confirmations.
    

    Methods
    -------
    None
    
    '''

    def __init__(self,
                 sessionType : str,
                 accountID : str,
                 token : str,
                 errorLog : str,
                 errorPrint : bool,
                 orderLog : str,
                 orderPrint : bool,
                 streamBeats : int,
                 streamRetries : int,
                 streamReset : int
                 ) -> None:
        ''' Initializes OandaSession object. 
        
        Parameters
        ----------
        `sessionType` : str
            Determines which oanda servers to send all subsequent communication
            to:
                "paper" = Paper account\n
                "live" = Live account

        `accountID` : str
            Unique Account ID for the account to trade with (identify
            within Oanda portal).
        
        `token` : str
            Unique token generated for Oanda account. *Note* All "live" accounts
            share the same token, but "paper" accounts have their own unique
            token - make sure you're providing the correct one for the 
            `sessionType` started.
        
        `errorLog` : str | None
            (Optional) Full path to log file on disk for recording errors. If
            provided, will attempt to load any pre-existing logs to memory
            before error logging begins.
        
        `errorPrint` : bool
            Whether to print errors to stdout.
        
        `orderLog` : str | None
            (Optional) Full path to log file on disk for recording confirmations. 
            If provided, will attempt to load any pre-existing logs to memory
            before confirmation logging begins.
        
        `orderPrint` : bool
            Whether to print order confirmations to stdout.
        
        `streamBeats` : int
            Number of seconds between heartbeats before a stream is considered dead.
        
        `streamRetries` : int
            Number of times to attempt to restart a dead stream.
        
        `streamReset` : int
            Number of minutes before resetting `streamRetries` counters back to zero for each endpoint.

        Returns
        -------
        `None`

        '''

        # initialize base configuration class
        self._base = session.BaseClient(sessionType=sessionType,
                                               accountID=accountID,
                                               token=token)

        # initialize all session classes
        self.account = session.Account(self._base)
        self.instruments = session.Instruments(self._base)
        self.orders = session.Orders(self._base)
        self.trades = session.Trades(self._base)
        self.positions = session.Positions(self._base)
        self.transactions = session.Transactions(self._base)
        self.pricing = session.Pricing(self._base)

        # initialize error monitoring
        self.errorMonitor = monitor.ErrorMonitor(endpoints=
                                                        [self.account, 
                                                         self.instruments, 
                                                         self.orders,                                                     
                                                         self.trades, 
                                                         self.positions, 
                                                         self.transactions,
                                                         self.pricing],
                                                         printErrors=errorPrint,
                                                         logPath=errorLog)
        self.errorMonitor.start()

        # initialize stream monitoring
        self.streamMonitor = monitor.StreamMonitor(endpoints=
                                                          [self.pricing, 
                                                           self.transactions],
                                                           deadOnArrival=streamBeats,
                                                           doNotResusitate=streamRetries,
                                                           resetTime=streamReset)
        self.streamMonitor.start()

        # initialize order monitoring
        self.orderMonitor = monitor.OrderMonitor(endpoints=
                                                        [self.orders,                                                     
                                                         self.trades, 
                                                         self.positions],
                                                         printConfirmations=orderPrint,
                                                         logPath=orderLog)
        self.orderMonitor.start()


        return None

    def quit(self) -> None:
        '''
        
        Gracefully stops the given session's sub-threads (monitors and streams),
        allowing the parent program to cleanly exit.
        

        Parameters
        ----------
        None

        Returns
        -------
        `None`

        '''

        # prevent streams from re-starting
        self.streamMonitor.doNotResusitate = 0

        # stop streams if started and still running
        if self.pricing._streamThread:
            if self.pricing._streamThread.is_alive():
                self.pricing.stop_stream()
        
        if self.transactions._streamThread:
            if self.transactions._streamThread.is_alive():
                self.transactions.stop_stream()
        
        # stop monitors
        self.streamMonitor.stop()
        self.errorMonitor.stop()
        self.orderMonitor.stop()

        return None

def start_session(sessionType : str,
                  accountID : str,
                  token : str,
                  errorLog : str | None = None,
                  errorPrint : bool = False,
                  orderLog : str | None = None,
                  orderPrint : bool = False,
                  streamBeats : int = 10,
                  streamRetries : int = 3,
                  streamReset : int = 60
                  ) -> None:
    '''  Instantiates an OandaSession object with API access to Oanda trading 
    endpoints.
    
    Parameters
    ----------
    `sessionType` : str
        Determines which oanda servers to send all subsequent communication
        to:
            sessionType="paper" : Paper account\n
            sessionType="live" : Live account

    `accountID` : str
        Unique Account ID for the account to trade with (identify
        within Oanda portal).
    
    `token` : str
        Unique token generated for Oanda account. *Note* All "live" accounts
        share the same token, but "paper" accounts have their own unique
        token - make sure you're providing the correct one for the 
        `sessionType` started.
    
    `errorLog` : str | None = None
        (Optional) Full path to log file on disk for recording errors. If
        provided, will attempt to load any pre-existing logs to memory
        before error logging begins. [Default=None]
    
    `errorPrint` : bool = False
        Whether to print errors to stdout. [Default=False]
    
    `orderLog` : str | None = None
        (Optional) Full path to log file on disk for recording confirmations. 
        If provided, will attempt to load any pre-existing logs to memory
        before confirmation logging begins. [Default=None]
    
    `orderPrint` : bool = False
        Whether to print order confirmations to stdout. [Default=False]
    
    `streamBeats` : int = 10
        Number of seconds between heartbeats before a stream is considered dead. [Default=10]
    
    `streamRetries` : int = 3
        Number of times to attempt to restart a dead stream. [Default=3]
    
    `streamReset` : int = 60
        Number of minutes before resetting `streamRetries` counters back to zero for each endpoint. [Default=60]
    
    Returns
    -------
    `OandaSession`
        Custom class object with API access to Oanda trading endpoints.
    
    '''
    
    # start Oanda session
    session = OandaSession(sessionType,
                            accountID,
                            token,
                            errorLog,
                            errorPrint,
                            orderLog,
                            orderPrint,
                            streamBeats,
                            streamRetries,
                            streamReset)

    return session

def to_baseUnits(currentQuotes : dict,
                 homeUnits : float,
                 truncate : bool = False) -> float | int:
    '''
    
    Convert units of the account's home currency to equivalent units of an 
    instrument's base currency.


    Parameters
    ----------
    `currentQuotes` : dict
        The current `session.pricing.pricing` details of the target
        instrument. *Note* Pricing data must include home conversion factors 
        - this is the default in session.pricing.update_pricing().
    
    `homeUnits` : float
        Units of the account's home currency to convert.

    `truncate` : bool = False
        Whether to truncate the equivalent units of the base currency. Set this
        value to `True` when calculating units for an order - OANDA order units 
        are the number of the target instrument's base currency that you'd like 
        to buy or sell - these units must be INTEGERS! When `truncate=True`, if the 
        equivalent units of a base currency contain decimals, the units will be 
        "floored" to the nearest integer (decimals will be dropped) to comply 
        with OANDA order specifications. This will result in an equivalent order 
        size that is slightly smaller than that requested in `homeUnits`. 
        To verify the true value of the base currency units after truncating, use 
        `easyoanda.calc_home()`. [default=False]

    Returns
    -------
    float | int
        The equivalent units of the target instrument's base currency.
    
    '''
    
    # pull conversion factor
    baseConversion = currentQuotes["homeConversions"][0]["positionValue"]

    # converting for an order
    if truncate:

        # units to buy / sell
        if homeUnits > 0:
            
            # floor if positive
            baseUnits = homeUnits // baseConversion

        else:

            # ceiling if negative
            baseUnits = -(-homeUnits // baseConversion)

        baseUnits = int(baseUnits)

    # else general conversion
    else:
        baseUnits = homeUnits / baseConversion

    return baseUnits

def to_homeUnits(currentQuotes : dict,
                 baseUnits : float | int) -> float:
    '''
    
    Convert units of an instrument's base currency to equivalent units of  
    the account's home currency.


    Parameters
    ----------
    `currentQuotes` : dict
        The current `session.pricing.pricing` details of the target
        instrument. *Note* Pricing data must include home conversion factors 
        - this is the default in session.pricing.update_pricing().
    
    `baseUnits` : float
        Units of the instrument's base currency to convert.

    Returns
    -------
    float
        The equivalent units of the account's home currency.
    
    '''

    # pull base conversion
    baseConversion = currentQuotes["homeConversions"][0]["positionValue"]

    # convert to home units
    homeUnits = baseUnits * baseConversion

    return homeUnits

def find_optimal_stop(currentQuotes : dict,
                      baseUnits : int,
                      maxLoss : float,
                      entryPrice : float | None = None) -> None:
    '''
    
    Calculates the optimal stop-loss price level given an order's units
    (quoted in the target instrument's base currency) and trader's 
    maximum loss threshold (quoted in the account's home currency). *Note*
    OANDA requires stop-loss price levels be rounded to their 5th decimal place - 
    this is an industry standard. Due to this rounding, potential losses from
    the optimal stop-loss price level are slightly smaller than those 
    requested in `maxLoss`. To verify the true value of potential losses in 
    the account's home currency, use `easyoanda.get_price_impact()`.

    
    Parameters
    ----------
    `currentQuotes` : dict
        The current `session.pricing.pricing` details of the target
        instrument. *Note* Pricing data must include home conversion factors 
        - this is the default in session.pricing.update_pricing().

    `baseUnits` : int
        The order size of the trade (quoted in the target instrument's base
        currency units). Positive units indicate a long position, negative 
        units indicate a short position. *Reminder* OANDA order units must be 
        INTEGERS.

    `maxLoss` : float
        The maximum allowable loss a trader is willing to take on the position
        (quoted in the account's home currency). 
    
    `entryPrice` : float | None = None
        The trade's projected entry price. If `None`, will assume trade is 
        a market order and will use most recently quoted bid / ask provided
        within `currentQuotes` (depending on sign of `baseUnits`). [default=None]

    Returns
    -------
    float
        The target instrument's optimal stop-loss price level.

    '''

    # pull quote conversion factor
    quoteConversion = currentQuotes["homeConversions"][1]["positionValue"]

    # per unit impact
    perUnitImpact = abs(baseUnits) * quoteConversion

    # distance = maxLoss / perUnitImpact
    distance = abs(maxLoss) / perUnitImpact

    # projected price already present
    if entryPrice:
        pass

    # or using current quotes
    else:
        # going long - setting price to most recent ask
        if baseUnits > 0:
            entryPrice = currentQuotes["prices"][0]["closeoutAsk"]

        # or going short - setting price to most recent bid
        else:
            entryPrice = currentQuotes["prices"][0]["closeoutBid"]


    # calculate stop for long
    if baseUnits > 0:
        stopLossAt = entryPrice - distance

        # round up to .0000X
        tempStopAt = stopLossAt * 100000
        tempStopAt = -(-tempStopAt // 1)
        stopLossAt = tempStopAt / 100000
    
    # calculate stop for short
    else:
        stopLossAt = entryPrice + distance

        # round down to .0000X
        tempStopAt = stopLossAt * 100000
        tempStopAt = tempStopAt // 1
        stopLossAt = tempStopAt / 100000


    return stopLossAt

def find_optimal_size(currentQuotes : dict,
                      maxLoss : float,
                      exitPrice : float,
                      entryPrice : float | None = None) -> None:

    '''

    Calculate the optimal order size for a trade (in the target instrument's base 
    currency), given a target stop-loss price level and trader's maximum loss 
    threshold (quoted in the account's home currency). *Note* OANDA order units 
    are the number of the target instrument's base currency that you'd like 
    to buy or sell - these units must be INTEGERS! After the optimal units
    are calculated, if they contain decimals, the units will be 
    "floored" to the nearest integer (decimals will be dropped) to comply 
    with OANDA order specifications. This will result in an order size that is 
    slightly less than optimal - a "best-fit", if you will. This "best-fit" size 
    is the closest to the optimal size while still keeping potential losses below 
    the trader's maximum loss threshold. To verify the true value of the 
    optimal order size in the account's home currency, use `easyoanda.calc_home()`.


    Parameters
    ----------
    `currentQuotes` : dict
        The current `session.pricing.pricing` details of the target
        instrument. *Note* Pricing data must include home conversion factors 
        - this is the default in session.pricing.update_pricing().

    `exitPrice` : float
        The trade's target stop-loss price level.

    `maxLoss` : float | None = None
        The maximum allowable loss a trader is willing to take on the position
        (quoted in the account's home currency).
    
    `entryPrice` : float | None = None
        The order's projected entry price. If `None`, will assume the order is 
        a market order and will use the most recently quoted bid / ask provided
        within `currentQuotes`. The average of the bid-ask is used as a 
        benchmark to evaluate the `exitPrice` against to determine if the
        position is long or short - if your market order stops are 
        extremely close to the bid/ask (anything less than half the spread), 
        it may be worthwhile to enter this parameter manually. [default=None]

    
    Returns
    -------
    int
        The optimal order size for the trade in the target instrument's base
        currency.

    '''

    # pull quote conversion factor
    quoteConversion = currentQuotes["homeConversions"][1]["positionValue"]

    # get entry price
    if entryPrice:
        pass
    else:
        
        # benchmark to determine if long or short
        benchmark = (currentQuotes["prices"][0]["closeoutAsk"]                 \
                   + currentQuotes["prices"][0]["closeoutBid"]) / 2


        # going short - setting price to most recent bid
        if exitPrice > benchmark:
            entryPrice = currentQuotes["prices"][0]["closeoutBid"]

        # or going long - setting price to most recent ask
        else:
            entryPrice = currentQuotes["prices"][0]["closeoutAsk"]


    # calculate distance betwen entry and exit loss
    distance = entryPrice - exitPrice

    # calculate target loss perUnitImpact
    lossPerUnitImpact = abs(maxLoss) / distance

    # if long
    if lossPerUnitImpact > 0:
        
        # floor if positive
        baseUnits = lossPerUnitImpact // quoteConversion

    else:

        # short - ceiling negative
        baseUnits = -(-lossPerUnitImpact // quoteConversion)

    return baseUnits

def get_pip_impact(currentQuotes : dict,
                   baseUnits : float) -> None:
    '''
    
    Calculate the price impact of a single pip change (as measured in the 
    account's home currency), given a number of units of the target instrument's 
    base currency. *Note* A "pip" for instrumented quoted in "JPY" or "HUF" is 
    .01, whereas for all others, a "pip" is .0001.

    
    Parameters
    ----------
    `currentQuotes` : dict
        The current `session.pricing.pricing` details of the target
        instrument. *Note* Pricing data must include home conversion factors 
        - this is the default in session.pricing.update_pricing().
    
    `baseUnits` : float
        Units of the instrument's base currency.

    Returns
    -------
    float
        The price impact a single pip change has (as measured in the 
        account's home currency)
    
    '''

    # pull quote conversion factor
    quoteConversion = currentQuotes["homeConversions"][1]["positionValue"]

    # calculating pip impact
    quotedUnits = baseUnits * quoteConversion

    # special pip adjustment if quoted in "JPY" or "HUF"
    if (currentQuotes["homeConversions"][1]["currency"] == "JPY") \
        or (currentQuotes["homeConversions"][1]["currency"] == "HUF"):
        perPipImpact = quotedUnits / 100

    # otherwise, standard pip adjustment
    else:
        perPipImpact = quotedUnits / 10000

    return abs(perPipImpact)

def get_price_impact(currentQuotes : dict,
                      baseUnits : float,
                      exitPrice : float,
                      entryPrice : float | None = None) -> None:

    '''
    
    Calculate the price impact of movements between two price levels within an 
    instrument (as measured in the account's home currency), given a number of 
    units of the target instrument's base currency.
    
    
    Parameters
    ----------
    `currentQuotes` : dict
        The current `session.pricing.pricing` details of the target
        instrument. *Note* Pricing data must include home conversion factors 
        - this is the default in session.pricing.update_pricing().

    `baseUnits` : float
        Units of the instrument's base currency. Positive for long position,
        negative for short position.
            
    `exitPrice` : float
        The instrument's ending price level.

    `entryPrice` : float | None = None
        The instrument's starting price level. If `None`, will assume entry
        price level is based on current bid/ask quotes (evaluated by sign of 
        `baseUnits`). [default=None]

    Returns
    -------
    float
        The price impact of changes between the two price levels (as measured
        in the account's home currency).

    '''


    # calculate pip impact
    pipImpact = get_pip_impact(currentQuotes, baseUnits)

    # went long - entered at the ask
    if baseUnits > 0:

        # set entry level as needed
        if entryPrice:
            pass
        else:
            entryPrice = currentQuotes["prices"][0]["closeoutAsk"]

        # calculate long distance
        distance = exitPrice - entryPrice
        
    # went short - entered at the bid
    else:

        # set entry level as needed
        if entryPrice:
            pass

        else:
            entryPrice = currentQuotes["prices"][0]["closeoutBid"]

        # calculate short distance
        distance = entryPrice - exitPrice

    # calculate pips in distance if "JPY" or "HUF"
    if (currentQuotes["homeConversions"][1]["currency"] == "JPY") \
        or (currentQuotes["homeConversions"][1]["currency"] == "HUF"):
        pips = distance * 100

    # otherwise, calculate standard pips
    else:
        pips = distance * 10000

    # calculate total position impact
    positionImpact = pipImpact * pips

    return positionImpact


















import requests
import threading
import json
import datetime
import copy

''' CLIENT-SERVER COMMUNICATIONS '''
def _to_objects(dictionary) -> dict:
    ''' Converts eligible string values to python datatypes (does NOT throw 
    errors on any failed conversions, value will just remain a string). Used as 
    an argument for json.loads() object hook conversions: 
    json.loads(<data>, object_hook=_to_objects). Supported conversion:
            int\n
            float\n
            datetime.datetime
    
    Parameters
    ----------
    dictionary: dict
        The json to iterate over.

    Returns
    -------
    dict
    
    '''
    for key in dictionary:
        if type(dictionary[key]) == str:
            
            # likely a float or RCF3339 time if string contains "."
            if "." in dictionary[key]:
                # more often than not, it's a float
                try: 
                    dictionary[key] = float(dictionary[key])
                except:
                    # otherwise typically RCF3339 time
                    try:
                        dictionary[key] = datetime.datetime.fromisoformat(dictionary[key])
                    # if neither, should likely remain a string
                    except:
                        pass
            
            # try converting integers, as well
            else:
                try: 
                    dictionary[key] = int(dictionary[key])
                except:
                    pass

    return dictionary

def _to_strings(obj : dict) -> dict:
    ''' Designed for pre-formatting server requests, recursively replaces objects
    with their  string equivalents (datetime.datetime objects are formatted as 
    RCF3339 in UTC). *Note* Times will be converted to UTC prior to conversion -
    ensure timezones are properly assigned within datetime objects if operating 
    in a different timezone.
    
    Parameters
    ----------
    obj : dict
        The dictionary to recurse over.

    Returns
    -------
    dict
        A dictionary with all eligible values (recursively) formatted as 
        strings (or None).

    '''

    # recurse down list
    if isinstance(obj, list):
        for entry in range(0, len(obj)):
            obj[entry] = _to_strings(obj[entry])

    # recurse down dictionary
    elif isinstance(obj, dict):
        for key in obj:
            obj[key] = _to_strings(obj[key])

    # check if object is a datetime (and convert / format if it is)
    elif isinstance(obj, datetime.datetime):
        obj = obj.astimezone(datetime.UTC).isoformat().replace("+00:00", "Z")
        if obj[-1] != "Z":
            obj = obj + "Z"
    
    # pass over None(s)
    elif obj is None:
        pass

    # otherwise, check if object is a string (and convert if not)
    elif not isinstance(obj, str):
        obj = str(obj)

    return obj

def get_request(endpoint : str,
                headers : None | dict = None,
                parameters : None | dict = None) -> tuple[dict, int]:
    ''' `requests.get()` wrapper with custom error handling.

    Parameters
    ----------------
    `endpoint` : str = ""
        Target URL.
        
    `header` : None | dict = None
        Headers to pass. 

    `parameters` : None | dict = None
        Parameters to pass.


    Returns
    -----------
    dict
        Dictionary of response.

    int
        Request response return code.

    '''

    # _to_strings() converts object's items in place
    cparameters = copy.deepcopy(parameters)

    # send request and validate response
    try:
        response = requests.get(endpoint, 
                            headers=headers, 
                            params=_to_strings(cparameters))
        status_code = response.status_code

    # catches all errors EXCEPT HTTPErrors
    except Exception as err:
        error = str(err)
        
        # arbitrary "non-HTTP error" status code
        status_code = 999

    # if success
    if (status_code == 200) or (status_code == 201):
        rdata = response.json(object_hook=_to_objects)

    # otherwise error occured
    else:
        # if non-HTTP error, error message already set
        if status_code == 999:
            pass
        # otherwise set HTTP error message
        else:
            error = response.json(object_hook=_to_objects)
        
        # build error message payload
        rdata = {"url" : endpoint,
                "headers" : copy.copy(headers),
                "parameters" : cparameters,
                "payload" : None,
                "code" : status_code,
                "message" : error}
        
        # remove Oanda token
        rdata["headers"]["Authorization"] = "<REDACTED>"


    return rdata, status_code

def post_request(endpoint : str,
                headers : None | dict = None,
                parameters : None | dict = None,
                data : None | dict = None) -> tuple[dict, int]:
    ''' `requests.post()` wrapper with custom error handling.

    Parameters
    ----------------
    `endpoint` : str = ""
        Target URL.
        
    `header` : None | dict = None
        Headers to pass. 

    `parameters` : None | dict = None
        Parameters to pass.

    `data` : None | dict = None
        Data to pass.

    Returns
    -----------
    dict
        Dictionary of response.

    int
        Request response return code.

    '''

    # _to_strings() converts object's items in place
    cparameters = copy.deepcopy(parameters)
    cdata = copy.deepcopy(data)
    
    # send request and validate response
    try:       
        response = requests.post(url=endpoint,
                        headers=headers, 
                        params=_to_strings(cparameters), 
                        json=_to_strings(cdata))
        status_code = response.status_code

    # catches all errors EXCEPT HTTPErrors
    except Exception as err:
        error = str(err)
        
        # arbitrary "non-HTTP error" status code
        status_code = 999

    # if success
    if (status_code == 200) or (status_code == 201):
        rdata = response.json(object_hook=_to_objects)

    # otherwise error occured
    else:
        # if non-HTTP error, error message already set
        if status_code == 999:
            pass
        # otherwise set HTTP error message
        else:
            error = response.json(object_hook=_to_objects)
        
        # build error message payload
        rdata = {"url" : endpoint,
                "headers" : copy.copy(headers),
                "parameters" : cparameters,
                "payload" : cdata,
                "code" : status_code,
                "message" : error}
        
        # remove Oanda token
        rdata["headers"]["Authorization"] = "<REDACTED>"

    return rdata, status_code

def put_request(endpoint : str,
                headers : None | dict = None,
                parameters : None | dict = None,
                data : None | dict = None) -> tuple[dict, int]:
    ''' Simple wrapper for requests.put()

    Parameters
    ----------------
    `endpoint` : str = ""
        Target URL.
        
    `header` : None | dict = None
        Headers to pass. 

    `parameters` : None | dict = None
        Parameters to pass.

    `data` : None | dict = None
        Data to pass.

    Returns
    -----------
    dict
        Dictionary of response.

    int
        Request response return code.

    '''

    # _to_strings() converts object's items in place
    cparameters = copy.deepcopy(parameters)
    cdata = copy.deepcopy(data)

    # send request and validate response
    try:
        response = requests.put(url=endpoint,
                            headers=headers, 
                            params=_to_strings(cparameters), 
                            json=_to_strings(cdata))
        status_code = response.status_code

    # catches all errors EXCEPT HTTPErrors
    except Exception as err:
        error = str(err)
        
        # arbitrary "non-HTTP error" status code
        status_code = 999

    # if success
    if (status_code == 200) or (status_code == 201):
        rdata = response.json(object_hook=_to_objects)

    # otherwise error occured
    else:
        # if non-HTTP error, error message already set
        if status_code == 999:
            pass
        # otherwise set HTTP error message
        else:
            error = response.json(object_hook=_to_objects)
        
        # build error message payload
        rdata = {"url" : endpoint,
                "headers" : copy.copy(headers),
                "parameters" : cparameters,
                "payload" : cdata,
                "code" : status_code,
                "message" : error}
        
        # remove Oanda token
        rdata["headers"]["Authorization"] = "<REDACTED>"

    return rdata, status_code

def patch_request(endpoint : str,
                headers : None | dict = None,
                parameters : None | dict = None,
                data : None | dict = None) -> tuple[dict, int]:
    ''' Simple wrapper for requests.patch()

    Parameters
    ----------------
    `endpoint` : str = ""
        Target URL.
        
    `header` : None | dict = None
        Headers to pass. 

    `parameters` : None | dict = None
        Parameters to pass.

    `data` : None | dict = None
        Data to pass.

    Returns
    -----------
    dict
        Dictionary of response.

    int
        Request response return code.

    '''

    # _to_strings() converts object's items in place
    cparameters = copy.deepcopy(parameters)
    cdata = copy.deepcopy(data)

    # send request and validate response
    try:        
        response = requests.patch(url=endpoint,
                              headers=headers, 
                              params=_to_strings(cparameters), 
                              json=_to_strings(cdata))
        status_code = response.status_code

    # catches all errors EXCEPT HTTPErrors
    except Exception as err:
        error = str(err)
        
        # arbitrary "non-HTTP error" status code
        status_code = 999

    # if success
    if (status_code == 200) or (status_code == 201):
        rdata = response.json(object_hook=_to_objects)

    # otherwise error occured
    else:
        # if non-HTTP error, error message already set
        if status_code == 999:
            pass
        # otherwise set HTTP error message
        else:
            error = response.json(object_hook=_to_objects)
        
        # build error message payload
        rdata = {"url" : endpoint,
                "headers" : copy.copy(headers),
                "parameters" : cparameters,
                "payload" : cdata,
                "code" : status_code,
                "message" : error}
        
        # remove Oanda token
        rdata["headers"]["Authorization"] = "<REDACTED>"

    return rdata, status_code

class _StreamingThread(threading.Thread):
    ''' Creates a thread to keep streamed content updated.

    Attributes
    ----------------
    `response` : requests.models.Response | None
        If the endpoint is succesfully reached (even if HTTPError), this will 
        contain the response generated from `requests.get(..., stream=True)`.

    `status_code` : int
        Either (1) the return code from the `requests.get(..., stream=True)` if
        the endpoint was reached (may or may not be an HTTPError code), or (2)
        an arbitrary `999` status code indicating some other error (network,
        programattic, etc).

    `errorMessage` : dict
        Error message details if `response.get(..., stream=True)` failed in
        any way (whether endpoint was reached or not):
            Keys:
            "url" : endpoint that received the request
            "headers" : headers sent to the endpoint (auth token is removed for 
                security purposes, check your token independently if you suspect 
                this is the issue)
            "parameters" : parameters sent to the endpoint
            "payload" : `None` for "GET" requests
            "code" : status code (HTTP code or `999`)
            "message" : error message

    `lastHeartbeat` : list
        If the stream is succesfully started, contains the most recent heartbeat 
        received from the stream (used to monitor stream's status).

    `content` : list
        If the stream is succesfully started and thread is recording content 
        (set by initialization parameter `record=True`),
        contains all received content since beginning of the stream, stored
        as a `list` (does not included stream "heartbeats") - this has the potential 
        to get very large if the stream runs for a while (be mindful of
        memory usage). If not recording content (set by initialization parameter
        `record=False`) contains each new single stream entry, stored as a `list`
        with a single entry - each new entry overwrites whatever previous single 
        entry was previously received (memory friendly).

    `_feed` : generator
        The stream's `self.iter_lines()` iterator object - `self.run()` iterates
        of this for updating content if the stream is succesfully started.

    `_recording` : bool
        (Flag) when True, appends new stream entries. When False, overwrites
        prior stream entry.

    `_stop_signal` : threading.Event
        When set, stops the given thread.

    Methods
    --------------
    `stop()` : func
        Sets `self._stop_signal`, stopping the given thread.


    `start()` : func
        Starts the thread (using `run()` function definition).

    `run()` : func
        Overrides threading.Thread.run(). Loops over `_feed` iterator object,
        appending/replacing new entries to `content`. Checks to see if 
        `self._stop_signal` is set before each new loop. Use `start()` to
        begin thread, not this function.
    
    Inheritance
    ----------------
    threading.Thread : class
        https://docs.python.org/3/library/threading.html#threading.Thread
    
    '''

    def __init__(self,
                 endpoint : str = "",
                 headers : None | dict = None,
                 parameters : None | dict = None,
                 record : bool = False) -> None:
        ''' Initializes StreamingThread object, begins connection stream.

        Parameters
        ----------------
        `endpoint` : str = ""
            Target URL.
            
        `header` : None | dict = None
            Headers to pass. 

        `parameters` : None | dict = None
            Parameters to pass.

        `record` : bool = False
            When set to True, the streamThread will record every new entry 
            it receives, appending the entry to the streamThread's `self.content` 
            attribute (as a list). If set to False, the streamThread will 
            overwrite it's `self.content` attribute with each new entry it 
            receives (as a dict). [Default=False]

        Returns
        -----------
        None

        '''

        # inherit Thread attributes
        threading.Thread.__init__(self)
        
        # create a way to stop the thread if need be. 
        self._stop_signal = threading.Event()
        
        # _to_strings() converts object's items in place
        cparameters = copy.deepcopy(parameters)

        # send request and validate response
        try:
            # start the stream
            self.response = requests.get(endpoint,
                            headers=headers,
                            params=_to_strings(cparameters),
                            stream=True)

            # set encoding type (how to translate octet-stream)
            if self.response.encoding is None:
                self.response.encoding = 'utf-8'

            # read in feed (decoding octet stream with above encoding type)
            self._feed = self.response.iter_lines(decode_unicode=True)
            
            # set status code
            self.status_code = self.response.status_code

            # set content container to record stream feed
            self.content = []
            self.lastHeartbeat = []
            self._recording = record

        # catches all errors EXCEPT HTTPErrors
        except Exception as err:
            
            # set error message to exception received
            error = str(err)
            
            # arbitrary "non-HTTP error" status code
            self.status_code = 999

        # if success, let thread run
        if (self.status_code == 200) or (self.status_code == 201):
            pass

        # otherwise error occured
        else:
            # if non-HTTP error, set null attributes (error message already set)
            if self.status_code == 999:
                self.response = None

            # otherwise set HTTP error message
            else:
                error = self.response.json(object_hook=_to_objects)
            
            # set additional null attributes
            self.lastHeartbeat = None
            self.content = None
            self._feed = None
            self._recording = None

            # kill thread even if accidentally started
            self._stop_signal.set()

            # build error message payload
            self.errorMessage = {"url" : endpoint,
                                 "headers" : copy.copy(headers),
                                 "parameters" : cparameters,
                                 "payload" : None,
                                 "code" : self.status_code,
                                 "message" : error}
            
            # remove Oanda token
            self.errorMessage["headers"]["Authorization"] = "<REDACTED>"

        return None

    def stop(self) -> None:
        ''' Stops the given streaming thread.

        Parameters
        ----------------
        None


        Returns
        -----------
        None

        '''
        self._stop_signal.set()

        return None

    def run(self) -> None:
        ''' Overrides threading.Thread.run(). Loops over `_feed` iterator object
        (the streamed content), and appends new entries to `content` if
        `recording=True`. Will continue to iterate until `self.stop()` is called.

        Parameters
        ----------------
        None


        Returns
        -----------
        None

        '''

        # read in feed line by line
        for line in self._feed:

            # keep doing this as long as thread is alive
            if not self._stop_signal.is_set():

                # if there's data to read
                if line:
                    # if it's a "HEARTBEAT", update the stream status:
                    if "HEARTBEAT" in line:
                        # convert strings to a dictionaries with proper objects
                        new_entry = json.loads(line, object_hook=_to_objects)

                        # if first heartbeat
                        if len(self.lastHeartbeat) == 0:
                            self.lastHeartbeat.append(new_entry)
                        # otherwise overwrite the last heartbeat
                        else:
                            self.lastHeartbeat[-1] = new_entry

                    # if it's not a "HEARTBEAT":
                    else:
                        # convert strings to a dictionaries with proper objects
                        new_entry = json.loads(line, object_hook=_to_objects)

                        # record entry if recording
                        if self._recording:
                            self.content.append(new_entry)
                        
                        # otherwise overwrite previous entry
                        else:
                            if len(self.content) == 0:
                                self.content.append(new_entry)
                            else:
                                self.content[-1] = new_entry
            
            # stop thread if `self._stop_signal` gets set via self.stop()
            else:
                break

def get_streamThread(endpoint : str,
                headers : None | dict = None,
                parameters : None | dict = None,
                record : bool = False) -> _StreamingThread:
    ''' Create a thread that collects and stores streamed content. 
    Thread must be started with self.start() once object is returned.

    Parameters
    ----------------
    `endpoint` : str = ""
        Target URL.
        
    `header` : None | dict = None
        Headers to pass. 

    `parameters` : None | dict = None
        Parameters to pass.

    `record` : bool = False
        When set to True, the streamThread will record every new entry it receives,
        appending the entry to the streamThread's `self.content` attribute (as a `list`).
        If set to False, the streamThread will overwrite it's `self.content` attribute
        with each new entry it receives (as a single-entry `list`). [Default=False]


    Returns
    -----------
    _StreamingThread
        Thread of the program managing the stream.

    '''
    streamThread = _StreamingThread(endpoint, 
                                   headers=headers, 
                                   parameters=parameters,
                                   record=record)

    return streamThread




import datetime

''' ENTRY ORDERS'''
class _BaseEntry():
    ''' Base entry order specifications. All entry orders extend this class.
    
    Attributes
    ----------
    `payload` : dict
        Specifications of the given entry order.

    Methods
    -------
    `set_takeProfit()` : func
        Creates and sets entry order's TakeProfit dependent order.
    
    `set_stopLoss()` : func
        Creates and sets entry order's StopLoss dependent order.
    
    `set_trailingStop()` : func
        Creates and sets entry order's TrailingStopLoss dependent order.
    
    `set_guaranteedStop()` : func
        Creates and sets entry order's GuarnateedStopLoss dependent order.
    
    `get_payload()` : func
        Returns entry order's payload.

    '''

    def __init__(self) -> None:
        ''' Initializes BaseOrder object. 

        Parameters
        ----------
        None

        Returns
        -------
        `None`
        
        '''

        # shared order specifications
        self.payload = {"type" : None,
                        "instrument" : None,
                        "units" : None,
                        "timeInForce" : None,
                        "positionFill" : None}

        return None
    
    def set_takeProfit(self,
                       price : float | None = None,
                       distance : float | None = None,
                       timeInForce : str = "GTC",
                       gtdTime : datetime.datetime | str | None = None) -> None:
        ''' Creates and sets entry order's TakeProfit dependent order.

        Parameters
        ----------
        `price` : float | None = None
            The associated Trade will be closed by a market price that is equal 
            to or better than this threshold (acts as Limit Order). Only 
            `price` OR `distance` may be specified - if both are input, 
            `price` will be given preference.

        `distance` : float | None = None
            Specifies the distance (in positive price units) from the trade's current 
            price to use as the Order price. The associated Trade will be closed
            by a market price that is equal to or better than this threshold 
            (acts as Limit Order). If the Trade is short the Instruments BID 
            price is used to calculated the price (and filled once ASK hits it), and 
            for long Trades the ASK is used (and filled once BID hits it). Only 
            `price` OR `distance` may be specified - if both are input, `price` 
            will be given preference.
        
        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce 
            describes how long an Order should remain pending before automaticaly 
            being cancelled by the execution system. Restricted to
            “GTC”, “GFD” and “GTD” for TakeProfit Orders [Default=GTC]:
        
            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce=GTD) The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.
        

        Returns
        -------
        `None`
        
        '''
        
        # create dependent field
        self.payload["takeProfitOnFill"] = {}

        # set required specifications
        if price:
            self.payload["takeProfitOnFill"]["price"] = price
        else:
            self.payload["takeProfitOnFill"]["distance"] = distance
        
        self.payload["takeProfitOnFill"]["timeInForce"] = timeInForce

        if (timeInForce == "GTD") and (gtdTime):
            self.payload["takeProfitOnFill"]["gtdTime"] = gtdTime

        return None
    
    def set_stopLoss(self,
                     price : float | None = None,
                     distance : float | None = None,
                     timeInForce : str = "GTC",
                     gtdTime : datetime.datetime | str | None = None) -> None:
        ''' Creates and sets entry order's StopLoss dependent order.

        Parameters
        ----------
        `price` : float | None = None
            The associated Trade will be closed by a market price that is equal 
            to or worse than this threshold (acts as Stop Order). Only 
            `price` OR `distance` may be specified - if both are input, 
            `price` will be given preference.

        `distance` : float | None = None
            Specifies the distance (in positive price units) from the trade's current 
            price to use as the Order price. The associated Trade will be closed
            by a market price that is equal to or worse than this threshold 
            (acts as Stop Order). If the Trade is short the Instruments BID 
            price is used to calculated the price (and filled once ASK hits it), and 
            for long Trades the ASK is used (and filled once BID hits it). Only 
            `price` OR `distance` may be specified - if both are input, `price` 
            will be given preference.
        
        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce 
            describes how long an Order should remain pending before automaticaly 
            being cancelled by the execution system. Restricted to
            “GTC”, “GFD” and “GTD” for StopLoss Orders [Default=GTC]:
        
            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce=GTD) The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.
        

        Returns
        -------
        `None`
        
        '''
        
        # create dependent field
        self.payload["stopLossOnFill"] = {}

        # set required specifications
        if price:
            self.payload["stopLossOnFill"]["price"] = price
        else:
            self.payload["stopLossOnFill"]["distance"] = distance
        
        self.payload["stopLossOnFill"]["timeInForce"] = timeInForce

        if (timeInForce == "GTD") and (gtdTime):
            self.payload["stopLossOnFill"]["gtdTime"] = gtdTime

        return None
    
    def set_trailingStop(self,
                         distance : float,
                         timeInForce : str = "GTC",
                         gtdTime : datetime.datetime | str | None = None) -> None:
        ''' Creates and sets entry order's TrailingStopLoss dependent order.

        Parameters
        ----------
        `distance` : float | None = None
            Specifies the distance (in positive price units) from the trade's current 
            price to use as the Order price. The associated Trade will be closed
            by a market price that is equal to or worse than this threshold 
            (acts as Stop Order). If the Trade is short the Instruments BID 
            price is used to calculated the price (and filled once ASK hits it), and 
            for long Trades the ASK is used (and filled once BID hits it).
        
        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce 
            describes how long an Order should remain pending before automaticaly 
            being cancelled by the execution system. Restricted to
            “GTC”, “GFD” and “GTD” for TrailingStopLoss Orders [Default=GTC]:
        
            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce=GTD) The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.
        

        Returns
        -------
        `None`
        
        '''

        # create dependent field
        self.payload["trailingStopLossOnFill"] = {}

        # set required specifications
        self.payload["trailingStopLossOnFill"]["distance"] = distance

        self.payload["trailingStopLossOnFill"]["timeInForce"] = timeInForce

        if (timeInForce == "GTD") and (gtdTime):
            self.payload["trailingStopLossOnFill"]["gtdTime"] = gtdTime

        return None
    
    def set_guaranteedStop(self,
                           price : float | None = None,
                           distance : float | None = None,
                           timeInForce : str = "GTC",
                           gtdTime : datetime.datetime | str | None = None) -> None:
        ''' Creates and sets entry order's GuarnateedStopLoss dependent order.

        Parameters
        ----------
        `price` : float | None = None
            The associated Trade will be closed at this price. Only 
            `price` OR `distance` may be specified - if both are input, 
            `price` will be given preference.

        `distance` : float | None = None
            Specifies the distance (in positive price units) from the trade's current 
            price to use as the Order price. The associated Trade will be closed
            at this price. If the Trade is short the Instruments BID 
            price is used to calculated the price (and filled once ASK hits it), and 
            for long Trades the ASK is used (and filled once BID hits it). Only 
            `price` OR `distance` may be specified - if both are input, `price` 
            will be given preference.
        
        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce 
            describes how long an Order should remain pending before automaticaly 
            being cancelled by the execution system. Restricted to
            “GTC”, “GFD” and “GTD” for GuarnateedStopLoss Orders [Default=GTC]:
        
            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce=GTD) The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.
        

        Returns
        -------
        `None`
        
        '''
        
        # create dependent field
        self.payload["guaranteedStopLossOnFill"] = {}

        # set required specifications
        if price:
            self.payload["guaranteedStopLossOnFill"]["price"] = price
        else:
            self.payload["guaranteedStopLossOnFill"]["distance"] = distance
        
        self.payload["guaranteedStopLossOnFill"]["timeInForce"] = timeInForce

        if (timeInForce == "GTD") and (gtdTime):
            self.payload["guaranteedStopLossOnFill"]["gtdTime"] = gtdTime

        return None

    def get_payload(self):
        ''' Returns entry order's payload.
        
        Parameters
        ----------
        None

        Returns
        -------
        dict
            The Oanda formatted entry order's specifications.
        
        '''

        return self.payload

class MarketOrder(_BaseEntry):
    ''' Market order specifications.
    
    Attributes
    ----------
    `payload` : dict
        Specifications of the given market order.

    Methods
    -------
    `set()` : func
        Sets required Market Order specifications. 

    '''

    def __init__(self) -> None:
        ''' Initializes MarketOrder object. 

        Parameters
        ----------
        None
        
        Returns
        -------
        `None`
        
        '''

        _BaseEntry.__init__(self)

        # set Market Order type
        self.payload["type"] = "MARKET"

        return None

    def set(self,
                  instrument : str,
                  units : int,
                  priceBounds : float | None = None,
                  timeInForce : str = "FOK",
                  positionFill : str = "DEFAULT") -> None:
        ''' Sets required Market Order specifications. 
        
        Parameters
        ----------
        `instrument` : str
            The order's target instrument.

        `units` : int
            The quantity requested to be filled by the order. A positive
            number of units results in a long Order, and a negative number of units
            results in a short Order.

        `priceBound` : float | None = None
            (Optional) The worst price that the client is willing to have the Order
            filled at.

        `timeInForce` : str = "FOK"
            The time-in-force requested for the Order. TimeInForce describes
            how long an Order should remain pending before automaticaly being
            cancelled by the execution system. Must be "FOK" or "IOC" for
            Market Orders [Default=FOK]:

            "FOK"	: The Order must be immediately “Filled Or Killed”\n
            "IOC"	: The Order must be “Immediately partially filled Or Cancelled”

        `positionFill` : str = "DEFAULT"
            Specification of how Positions in the Account are modified when the Order
            is filled [Default=DEFAULT]:

            "OPEN_ONLY"	: When the Order is filled, only allow Positions to be 
                opened or extended.
            "REDUCE_FIRST"	: When the Order is filled, always fully reduce an 
                existing Position before opening a new Position.
            "REDUCE_ONLY"	: When the Order is filled, only reduce an existing 
                Position.
            "DEFAULT"	: When the Order is filled, use REDUCE_FIRST behaviour 
                for non-client hedging Accounts, and OPEN_ONLY behaviour for 
                client hedging Accounts.

        Returns
        -------
        `None`

        '''

        
        # set required specifications
        self.payload["instrument"] = instrument
        self.payload["units"] = units
        self.payload["timeInForce"] = timeInForce
        self.payload["positionFill"] = positionFill
        
        # set optional specifications
        if priceBounds:
            self.payload["priceBounds"] = priceBounds

        return None
    
class LimitOrder(_BaseEntry):
    ''' Limit order specifications. *Note* A general note on LimitOrders:

    If POSITIVE units provided (Going Long / Closing Short)...
        AND Current Price < Order Price: 
            order will be filled immediately at CURRENT market prices (if not
            enough market liquidity and markets move UPWARD, will continue to be 
            filled only at prices LESS THAN or EQUAL TO the ORDER price)
        
        AND Current Price = Order Price: 
            order will be filled immediately at ORDER / CURRENT price or LESS 
            (if enough market liquidity)

        AND Current Price > Order Price: 
            order will sit at ORDER price until CURRENT price FALLS to ORDER price,
            at which point the order will be filled at ORDER price or LESS (if 
            enough market liquidity)
        
    If Negative Units Provided (Going Short / Closing Long) and...
        AND Current Price < Order Price: 
            order will sit at ORDER price until CURRENT price RISES to ORDER price,
            at which point the order will be filled at ORDER price or GREATER 
            (if enough market liquidity)

        AND Current Price = Order Price: 
            order will be filled immediately at ORDER / CURRENT price or GREATER
            (if enough market liquidity)
        
        AND Current Price > Order Price: 
            order will be filled immediately at CURRENT market prices (if not
            enough market liquidity and markets move DOWNWARD, will continue to
            be filled  only at prices GREATER THAN or EQUAL TO the ORDER price)
        
    
    Attributes
    ----------
    `payload` : dict
        Specifications of the given limit order.

    Methods
    -------
    `set()` : func
        Sets required Limit Order specifications. 

    '''

    def __init__(self) -> None:
        ''' Initializes LimitOrder object. 

        Parameters
        ----------
        None
        
        Returns
        -------
        `None`
        
        '''

        _BaseEntry.__init__(self)

        # set Market Order type
        self.payload["type"] = "LIMIT"

        return None

    def set(self,
                  instrument : str,
                  units : int,
                  price : float,
                  timeInForce : str = "GTC",
                  gtdTime : datetime.datetime | str | None = None,
                  positionFill : str = "DEFAULT",
                  triggerCondition : str = "DEFAULT") -> None:
        ''' Sets required Limit Order specifications. 

        Parameters
        ----------
        `instrument` : str
            The order's target instrument.

        `units` : int
            The quantity requested to be filled by the order. A positive
            number of units results in a long Order, and a negative number of units
            results in a short Order.

        `price` : float | None = None
            The price threshold specified for the Order. The Limit Order will 
            only be filled by a market price that is equal to or better than 
            this price.

        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce describes
            how long an Order should remain pending before automaticaly being
            cancelled by the execution system [Default=GTC]:

            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time
            "FOK"	: The Order must be immediately “Filled Or Killed”
            "IOC"	: The Order must be “Immediately partially filled Or Cancelled”

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce="GTD") The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.

        `positionFill` : str = "DEFAULT"
            Specification of how Positions in the Account are modified when the Order
            is filled [Default=DEFAULT]:

            "OPEN_ONLY"	: When the Order is filled, only allow Positions to be 
                opened or extended.
            "REDUCE_FIRST"	: When the Order is filled, always fully reduce an 
                existing Position before opening a new Position.
            "REDUCE_ONLY"	: When the Order is filled, only reduce an existing 
                Position.
            "DEFAULT"	: When the Order is filled, use REDUCE_FIRST behaviour 
                for non-client hedging Accounts, and OPEN_ONLY behaviour for 
                client hedging Accounts.

        `triggerCondition` : str = "DEFAULT"
            Specification of which price component should be evaluated when
            determining if an Order should be triggered and filled [Default=DEFAULT]. 

            "DEFAULT"	: Trigger an Order the “natural” way: compare its price 
                to the ask for long Orders and bid for short Orders.
            "INVERSE"	: Trigger an Order the opposite of the “natural” way: 
                compare its price the bid for long Orders and ask for short Orders.
            "BID"	: Trigger an Order by comparing its price to the bid 
                regardless of whether it is long or short.
            "ASK"	: Trigger an Order by comparing its price to the ask 
                regardless of whether it is long or short.
            "MID"	: Trigger an Order by comparing its price to the midpoint 
                regardless of whether it is long or short.

                
        Returns
        -------
        `None`

        '''

        
        # set required specifications
        self.payload["instrument"] = instrument
        self.payload["units"] = units
        self.payload["price"] = price
        self.payload["timeInForce"] = timeInForce
        
        if (timeInForce == "GTD") and (gtdTime):
            self.payload["gtdTime"] = gtdTime

        self.payload["positionFill"] = positionFill
        self.payload["triggerCondition"] = triggerCondition

        return None

class StopOrder(_BaseEntry):
    ''' Stop order specifications. *Note* A general note on StopOrders:

    If POSITIVE units provided (Going Long / Closing Short)...
        AND Current Price < Order Price: 
            order will sit at ORDER price until CURRENT price RISES to ORDER price,
            at which point the order will be filled at the ORDER price or
            GREATER (if enough market liquidity)

        AND Current Price = Order Price: 
            order will be filled immediately at ORDER / CURRENT price or GREATER
            (if enough market liquidity)
        
        AND Current Price > Order Price: 
            order will be filled immediately at CURRENT market prices (if not
            enough market liquidity and markets move DOWNWARD, will continue to
            be filled only at prices GREATER THAN or EQUAL TO the ORDER price).
        
    If Negative Units Provided (Going Short / Closing Long)...
        AND Current Price > Order Price:
            order will sit at ORDER price until CURRENT prices FALL to ORDER price,
            at which point the order will be filled at the ORDER price or LESS
            (if enough market liquidity)

        AND Current Price = Order Price: 
            order will be filled immediately at ORDER / CURRENT price or LESS
            (if enough market liquidity)
        
        AND Current Price < Order Price: 
            order will be filled immediately at CURRENT market prices (if not
            enough market liquidity and markets move UPWARD, will continue to
            be filled only at prices LESS THAN or EQUAL TO the ORDER price)
    
    Attributes
    ----------
    `payload` : dict
        Specifications of the given stop order.

    Methods
    -------
    `set()` : func
        Sets required Stop Order specifications. 

    '''

    def __init__(self) -> None:
        ''' Initializes StopOrder object. 

        Parameters
        ----------
        None
        
        Returns
        -------
        `None`
        
        '''

        _BaseEntry.__init__(self)

        # set Market Order type
        self.payload["type"] = "STOP"

        return None

    def set(self,
                  instrument : str,
                  units : int,
                  price : float,
                  priceBound : float | None = None,
                  timeInForce : str = "GTC",
                  gtdTime : datetime.datetime | str | None = None,
                  positionFill : str = "DEFAULT",
                  triggerCondition : str = "DEFAULT") -> None:
        ''' Sets required Stop Order specifications. 
        
        Parameters
        ----------
        `instrument` : str
            The order's target instrument.

        `units` : int
            The quantity requested to be filled by the order. A positive
            number of units results in a long Order, and a negative number of units
            results in a short Order.

        `price` : float
            The price threshold specified for the Order. The Stop Order will 
            only be filled by a market price that is equal to or worse than this 
            price.

        `priceBound` : float | None = None
            (Optional) The worst price that the client is willing to have the Order
            filled at.

        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce describes
            how long an Order should remain pending before automaticaly being
            cancelled by the execution system [Default=GTC]:

            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time
            "FOK"	: The Order must be immediately “Filled Or Killed”
            "IOC"	: The Order must be “Immediately partially filled Or Cancelled”

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce="GTD") The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.

        `positionFill` : str = "DEFAULT"
            Specification of how Positions in the Account are modified when the Order
            is filled [Default=DEFAULT]:

            "OPEN_ONLY"	: When the Order is filled, only allow Positions to be 
                opened or extended.
            "REDUCE_FIRST"	: When the Order is filled, always fully reduce an 
                existing Position before opening a new Position.
            "REDUCE_ONLY"	: When the Order is filled, only reduce an existing 
                Position.
            "DEFAULT"	: When the Order is filled, use REDUCE_FIRST behaviour 
                for non-client hedging Accounts, and OPEN_ONLY behaviour for 
                client hedging Accounts.

        `triggerCondition` : str = "DEFAULT"
            Specification of which price component should be evaluated when
            determining if an Order should be triggered and filled [Default=DEFAULT]. 

            "DEFAULT"	: Trigger an Order the “natural” way: compare its price 
                to the ask for long Orders and bid for short Orders.
            "INVERSE"	: Trigger an Order the opposite of the “natural” way: 
                compare its price the bid for long Orders and ask for short Orders.
            "BID"	: Trigger an Order by comparing its price to the bid 
                regardless of whether it is long or short.
            "ASK"	: Trigger an Order by comparing its price to the ask 
                regardless of whether it is long or short.
            "MID"	: Trigger an Order by comparing its price to the midpoint 
                regardless of whether it is long or short.

                
        Returns
        -------
        `None`

        '''

        
        # set required specifications
        self.payload["instrument"] = instrument
        self.payload["units"] = units
        self.payload["price"] = price
        self.payload["timeInForce"] = timeInForce
        
        if (timeInForce == "GTD") and (gtdTime):
            self.payload["gtdTime"] = gtdTime

        self.payload["positionFill"] = positionFill
        self.payload["triggerCondition"] = triggerCondition

        # set optional specifications
        if priceBound:
            self.payload["priceBound"] = priceBound

        return None

class MarketIfTouchedOrder(_BaseEntry):
    ''' MarketIfTouched order specifications. *Note* A general note on 
    MarketIfTouchedOrders:

    Think of a MarketIfTouchedOrder as taking ONE direction at a specific
    price point no matter where the market price comes from before hand.

    If POSITIVE units provided (Going Long / Closing Short)...
        AND Current Price < Order Price: 
            [Acts as Long Stop] order will sit at ORDER price until CURRENT price 
            RISES to ORDER price, at which point the order will be filled at the 
            ORDER price or GREATER (if enough market liquidity)

        AND Current Price = Order Price: 
            N/A
        
        AND Current Price > Order Price: 
            [Acts as Long Limit]  order will sit at ORDER price until CURRENT price 
            FALLS to ORDER price, at which point the order will be filled at 
            ORDER price or LESS (if enough market liquidity)
        
    If Negative Units Provided (Going Short / Closing Long)...
        AND Current Price > Order Price: 
            [Acts as Short Stop] order will sit at ORDER price until CURRENT price
            FALLS to ORDER price, at which point the order will be filled at the
            ORDER price or LESS (if enough market liquidity)

        AND Current Price = Order Price: 
            N/A

        AND Current Price < Order Price:
            [Acts as Short Limit] order will sit at ORDER price until CURRENT price 
            RISES to ORDER price, at which point the order will be filled at 
            ORDER price or GREATER (if enough market liquidity)
    
    Attributes
    ----------
    `payload` : dict
        Specifications of the given market-if-touched order.

    Methods
    -------
    `set()` : func
        Sets required MarketIfTouched Order specifications. 

    '''

    def __init__(self) -> None:
        ''' Initializes MarketIfTouchedOrder object. 

        Parameters
        ----------
        None
        
        Returns
        -------
        `None`
        
        '''

        _BaseEntry.__init__(self)

        # set Market Order type
        self.payload["type"] = "MARKET_IF_TOUCHED"

        return None

    def set(self,
                  instrument : str,
                  units : int,
                  price : float,
                  priceBound : float | None = None,
                  timeInForce : str = "GTC",
                  gtdTime : datetime.datetime | str | None = None,
                  positionFill : str = "DEFAULT",
                  triggerCondition : str = "DEFAULT") -> None:
        ''' Sets required MarketIfTouched Order specifications. 
        
        Parameters
        ----------
        `instrument` : str
            The order's target instrument.

        `units` : int
            The quantity requested to be filled by the order. A positive
            number of units results in a long Order, and a negative number of units
            results in a short Order.

        `price` : float
            The price threshold specified for the Order. The MarketIfTouched 
            Order will only be filled by a market price that crosses this price 
            from the direction of the market price at the time when the Order 
            was created (the initialMarketPrice). Depending on the value of the 
            Orders price and initialMarketPrice, the MarketIfTouchedOrder will 
            behave like a Limit or a Stop Order.

        `priceBound` : float | None = None
            (Optional) The worst price that the client is willing to have the Order
            filled at.

        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce describes
            how long an Order should remain pending before automaticaly being
            cancelled by the execution system. Restricted to “GTC”, “GFD” and 
            “GTD” for MarketIfTouched Orders [Default=GTC]:

            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce="GTD") The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.

        `positionFill` : str = "DEFAULT"
            Specification of how Positions in the Account are modified when the Order
            is filled [Default=DEFAULT]:

            "OPEN_ONLY"	: When the Order is filled, only allow Positions to be 
                opened or extended.
            "REDUCE_FIRST"	: When the Order is filled, always fully reduce an 
                existing Position before opening a new Position.
            "REDUCE_ONLY"	: When the Order is filled, only reduce an existing 
                Position.
            "DEFAULT"	: When the Order is filled, use REDUCE_FIRST behaviour 
                for non-client hedging Accounts, and OPEN_ONLY behaviour for 
                client hedging Accounts.

        `triggerCondition` : str = "DEFAULT"
            Specification of which price component should be evaluated when
            determining if an Order should be triggered and filled [Default=DEFAULT]. 

            "DEFAULT"	: Trigger an Order the “natural” way: compare its price 
                to the ask for long Orders and bid for short Orders.
            "INVERSE"	: Trigger an Order the opposite of the “natural” way: 
                compare its price the bid for long Orders and ask for short Orders.
            "BID"	: Trigger an Order by comparing its price to the bid 
                regardless of whether it is long or short.
            "ASK"	: Trigger an Order by comparing its price to the ask 
                regardless of whether it is long or short.
            "MID"	: Trigger an Order by comparing its price to the midpoint 
                regardless of whether it is long or short.

                
        Returns
        -------
        `None`

        '''

        
        # set required specifications
        self.payload["instrument"] = instrument
        self.payload["units"] = units
        self.payload["price"] = price
        self.payload["timeInForce"] = timeInForce
        
        if (timeInForce == "GTD") and (gtdTime):
            self.payload["gtdTime"] = gtdTime

        self.payload["positionFill"] = positionFill
        self.payload["triggerCondition"] = triggerCondition

        # set optional specifications
        if priceBound:
            self.payload["priceBound"] = priceBound

        return None

''' DEPENDENT ORDERS '''
class _BaseDependent():
    ''' Base dependent order specifications. All dependent orders extend this 
    class.
    
    Attributes
    ----------
    `payload` : dict
        Specifications of the given dependent order.

    Methods
    -------
    `get_payload()` : func
        Returns dependent order's payload.

    '''

    def __init__(self) -> None:
        ''' Initializes BaseDependent object. 

        Parameters
        ----------
        None

        Returns
        -------
        `None`
        
        '''

        # shared dependent specifications
        self.payload = {"type" : None,
                        "tradeID" : None,
                        "timeInForce" : None,
                        "triggerCondition": None}

        return None

    def get_payload(self):
        ''' Returns dependent order's payload.
        
        Parameters
        ----------
        None

        Returns
        -------
        dict
            The Oanda formatted dependent order's specifications.
        
        '''

        return self.payload

class TakeProfitOrder(_BaseDependent):
    ''' TakeProfit Order specifications.

    Attributes
    ----------
    `payload` : dict
        Specifications of the given TakeProfit order.

    Methods
    -------
    `set()` : func
        Sets required TakeProfit order specifications. 

    '''

    def __init__(self) -> None:
        ''' Instantiates TakeProfitOrder object.

        Parameters
        ----------
        None

        Returns
        -------
        `None`
        
        '''

        # inherit base class
        _BaseDependent.__init__(self)

        # set payload type
        self.payload["type"] = "TAKE_PROFIT"

        return None

    def set(self,
                   tradeID : int,
                   price : float | None = None,
                   distance : float | None = None,
                   timeInForce : str = "GTC",
                   gtdTime : datetime.datetime | str | None = None,
                   triggerCondition : str = "DEFAULT") -> None:
        ''' Sets required TakeProfit Order requirements.
        
        Parameters
        ----------
        `tradeID` : int
             The ID of the Trade to close when the price threshold is breached.

        `price` : float | None = None
            The associated Trade will be closed by a market price that is equal 
            to or better than this threshold (acts as Limit Order). Only 
            `price` OR `distance` may be specified - if both are input, 
            `price` will be given preference.

        `distance` : float | None = None
            Specifies the distance (in positive price units) from the trade's current 
            price to use as the Order price. The associated Trade will be closed
            by a market price that is equal to or better than this threshold 
            (acts as Limit Order). If the Trade is short the Instruments BID 
            price is used to calculated the price (and filled once ASK hits it), and 
            for long Trades the ASK is used (and filled once BID hits it). Only 
            `price` OR `distance` may be specified - if both are input, `price` 
            will be given preference.

        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce 
            describes how long an Order should remain pending before automaticaly 
            being cancelled by the execution system. Restricted to
            “GTC”, “GFD” and “GTD” for TakeProfit Orders [Default=GTC]:
        
            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce=GTD) The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.

        `triggerCondition` : str = "DEFAULT"
            Specification of which price component should be evaluated when
            determining if an Order should be triggered and filled [Default=DEFAULT]. 

            "DEFAULT"	: Trigger an Order the “natural” way: compare its price 
                to the ask for long Orders and bid for short Orders.
            "INVERSE"	: Trigger an Order the opposite of the “natural” way: 
                compare its price the bid for long Orders and ask for short Orders.
            "BID"	: Trigger an Order by comparing its price to the bid 
                regardless of whether it is long or short.
            "ASK"	: Trigger an Order by comparing its price to the ask 
                regardless of whether it is long or short.
            "MID"	: Trigger an Order by comparing its price to the midpoint 
                regardless of whether it is long or short.

        Returns
        -------
        `None`
        
        '''

        # set required specifications
        self.payload["tradeID"] = tradeID
        
        if price:
            self.payload["price"] = price
        else:
            self.payload["distance"] = distance
        
        self.payload["timeInForce"] = timeInForce

        if (timeInForce == "GTD") and (gtdTime):
            self.payload["gtdTime"] = gtdTime

        self.payload["triggerCondition"] = triggerCondition

        return None

class StopLossOrder(_BaseDependent):
    ''' StopLoss Order specifications.

    Attributes
    ----------
    `payload` : dict
        Specifications of the given StopLoss order.

    Methods
    -------
    `set()` : func
        Sets required StopLoss order specifications. 

    '''

    def __init__(self) -> None:
        ''' Instantiates StopLossOrder object.

        Parameters
        ----------
        None

        Returns
        -------
        `None`
        
        '''
        # inherit base class
        _BaseDependent.__init__(self)

        # set payload type
        self.payload["type"] = "STOP_LOSS"

        return None

    def set(self,
                   tradeID : int,
                   price : float | None = None,
                   distance : float | None = None,
                   timeInForce : str = "GTC",
                   gtdTime : datetime.datetime | str | None = None,
                   triggerCondition : str = "DEFAULT") -> None:
        ''' Sets required StopLoss Order requirements.
        
        Parameters
        ----------
        `tradeID` : int
             The ID of the Trade to close when the price threshold is breached.

        `price` : float | None = None
            The associated Trade will be closed by a market price that is equal 
            to or worse than this threshold (acts as Stop Order). Only 
            `price` OR `distance` may be specified - if both are input, 
            `price` will be given preference.

        `distance` : float | None = None
            Specifies the distance (in positive price units) from the trade's current 
            price to use as the Order price. The associated Trade will be closed
            by a market price that is equal to or better than this threshold 
            (acts as Limit Order). If the Trade is short the Instruments BID 
            price is used to calculated the price (and filled once ASK hits it), and 
            for long Trades the ASK is used (and filled once BID hits it). Only 
            `price` OR `distance` may be specified - if both are input, `price` 
            will be given preference.

        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce 
            describes how long an Order should remain pending before automaticaly 
            being cancelled by the execution system. Restricted to
            “GTC”, “GFD” and “GTD” for StopLoss Orders [Default=GTC]:
        
            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce=GTD) The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.

        `triggerCondition` : str = "DEFAULT"
            Specification of which price component should be evaluated when
            determining if an Order should be triggered and filled [Default=DEFAULT]. 

            "DEFAULT"	: Trigger an Order the “natural” way: compare its price 
                to the ask for long Orders and bid for short Orders.
            "INVERSE"	: Trigger an Order the opposite of the “natural” way: 
                compare its price the bid for long Orders and ask for short Orders.
            "BID"	: Trigger an Order by comparing its price to the bid 
                regardless of whether it is long or short.
            "ASK"	: Trigger an Order by comparing its price to the ask 
                regardless of whether it is long or short.
            "MID"	: Trigger an Order by comparing its price to the midpoint 
                regardless of whether it is long or short.

        Returns
        -------
        `None`
        
        '''

        # set required specifications
        self.payload["tradeID"] = tradeID
        
        if price:
            self.payload["price"] = price
        else:
            self.payload["distance"] = distance
        
        self.payload["timeInForce"] = timeInForce

        if (timeInForce == "GTD") and (gtdTime):
            self.payload["gtdTime"] = gtdTime

        self.payload["triggerCondition"] = triggerCondition

        return None

class GuaranteedStopLossOrder(_BaseDependent):
    ''' GuaranteedStopLoss Order specifications.

    Attributes
    ----------
    `payload` : dict
        Specifications of the given GuaranteedStopLoss order.

    Methods
    -------
    `set()` : func
        Sets required GuaranteedStopLoss order specifications. 

    '''

    def __init__(self) -> None:
        ''' Instantiates GuaranteedStopLossOrder object.

        Parameters
        ----------
        None

        Returns
        -------
        `None`
        
        '''

        # inherit base class
        _BaseDependent.__init__(self)

        # set payload type
        self.payload["type"] = "GUARANTEED_STOP_LOSS"

        return None

    def set(self,
                   tradeID : int,
                   price : float | None = None,
                   distance : float | None = None,
                   timeInForce : str = "GTC",
                   gtdTime : datetime.datetime | str | None = None,
                   triggerCondition : str = "DEFAULT") -> None:
        ''' Sets required GuaranteedStopLoss Order requirements.
        
        Parameters
        ----------
        `tradeID` : int
             The ID of the Trade to close when the price threshold is breached.

        `price` : float | None = None
            The associated Trade will be closed at this price. Only 
            `price` OR `distance` may be specified - if both are input, 
            `price` will be given preference.

        `distance` : float | None = None
            Specifies the distance (in positive price units) from the trade's current 
            price to use as the Order price. The associated Trade will be closed
            by a market price that is equal to or better than this threshold 
            (acts as Limit Order). If the Trade is short the Instruments BID 
            price is used to calculated the price (and filled once ASK hits it), and 
            for long Trades the ASK is used (and filled once BID hits it). Only 
            `price` OR `distance` may be specified - if both are input, `price` 
            will be given preference.

        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce 
            describes how long an Order should remain pending before automaticaly 
            being cancelled by the execution system. Restricted to
            “GTC”, “GFD” and “GTD” for GuaranteedStopLoss Orders [Default=GTC]:
        
            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce=GTD) The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.

        `triggerCondition` : str = "DEFAULT"
            Specification of which price component should be evaluated when
            determining if an Order should be triggered and filled [Default=DEFAULT]. 

            "DEFAULT"	: Trigger an Order the “natural” way: compare its price 
                to the ask for long Orders and bid for short Orders.
            "INVERSE"	: Trigger an Order the opposite of the “natural” way: 
                compare its price the bid for long Orders and ask for short Orders.
            "BID"	: Trigger an Order by comparing its price to the bid 
                regardless of whether it is long or short.
            "ASK"	: Trigger an Order by comparing its price to the ask 
                regardless of whether it is long or short.
            "MID"	: Trigger an Order by comparing its price to the midpoint 
                regardless of whether it is long or short.

        Returns
        -------
        `None`
        
        '''

        # set required specifications
        self.payload["tradeID"] = tradeID
        
        if price:
            self.payload["price"] = price
        else:
            self.payload["distance"] = distance
        
        self.payload["timeInForce"] = timeInForce

        if (timeInForce == "GTD") and (gtdTime):
            self.payload["gtdTime"] = gtdTime

        self.payload["triggerCondition"] = triggerCondition

        return None

class TrailingStopLossOrder(_BaseDependent):
    ''' TrailingStopLoss Order specifications.

    Attributes
    ----------
    `payload` : dict
        Specifications of the given TrailingStopLoss order.

    Methods
    -------
    `set()` : func
        Sets required TrailingStopLoss order specifications. 

    '''

    def __init__(self) -> None:
        ''' Instantiates TrailingStopLossOrder object.

        Parameters
        ----------
        None

        Returns
        -------
        `None`
        
        '''
        
        # inherit base class
        _BaseDependent.__init__(self)

        # set payload type
        self.payload["type"] = "TRAILING_STOP_LOSS"

        return None

    def set(self,
                   tradeID : int,
                   distance : float,
                   timeInForce : str = "GTC",
                   gtdTime : datetime.datetime | str | None = None,
                   triggerCondition : str = "DEFAULT") -> None:
        ''' Sets required TrailingStopLoss Order requirements.
        
        Parameters
        ----------
        `tradeID` : int
             The ID of the Trade to close when the price threshold is breached.
             
        `distance` : float
            Specifies the distance (in positive price units) from the trade's current 
            price to use as the Order price. The associated Trade will be closed
            by a market price that is equal to or worse than this threshold 
            (acts as Stop Order). If the Trade is short the Instruments BID 
            price is used to calculated the price (and filled once ASK hits it), and 
            for long Trades the ASK is used (and filled once BID hits it).

        `timeInForce` : str = "GTC"
            The time-in-force requested for the Order. TimeInForce 
            describes how long an Order should remain pending before automaticaly 
            being cancelled by the execution system. Restricted to
            “GTC”, “GFD” and “GTD” for GuaranteedStopLoss Orders [Default=GTC]:
        
            "GTC"	: The Order is “Good unTil Cancelled”
            "GTD"	: The Order is “Good unTil Date” and will be cancelled at 
                the provided time
            "GFD"	: The Order is “Good For Day” and will be cancelled at 5pm 
                New York time

        `gtdTime` : datetime.datetime | str | None = None
            (Required if timeInForce=GTD) The date/time when the Order will be 
            cancelled if its timeInForce is “GTD”. If string, ensure UTC in 
            RCF3339 formatted.

        `triggerCondition` : str = "DEFAULT"
            Specification of which price component should be evaluated when
            determining if an Order should be triggered and filled [Default=DEFAULT]. 

            "DEFAULT"	: Trigger an Order the “natural” way: compare its price 
                to the ask for long Orders and bid for short Orders.
            "INVERSE"	: Trigger an Order the opposite of the “natural” way: 
                compare its price the bid for long Orders and ask for short Orders.
            "BID"	: Trigger an Order by comparing its price to the bid 
                regardless of whether it is long or short.
            "ASK"	: Trigger an Order by comparing its price to the ask 
                regardless of whether it is long or short.
            "MID"	: Trigger an Order by comparing its price to the midpoint 
                regardless of whether it is long or short.

        Returns
        -------
        `None`
        
        '''

        # set required specifications
        self.payload["tradeID"] = tradeID
        
        self.payload["distance"] = distance
        
        self.payload["timeInForce"] = timeInForce

        if (timeInForce == "GTD") and (gtdTime):
            self.payload["gtdTime"] = gtdTime

        self.payload["triggerCondition"] = triggerCondition

        return None

